/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include "sort_machine.h"

class SortExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> prev_;
    ColMeta col_;                              // 框架中只支持一个键排序，需要自行修改数据结构支持多个键排序
    size_t tuple_num;
    bool is_desc_;
    bool isend=false;
    int tuple_len_;
    std::vector<size_t> used_tuple;
    std::unique_ptr<RmRecord> current_tuple;
    std::unique_ptr<sort_machine> sort_mac;
    std::ifstream readFile;
   public:
    SortExecutor(std::unique_ptr<AbstractExecutor> prev, TabCol sel_cols, bool is_desc) {
        prev_ = std::move(prev);
        for(auto col:prev_->cols()){
            if(col.name==sel_cols.col_name&&col.tab_name==sel_cols.tab_name){
                col_.len=col.len;
                col_.offset=col.offset;
                col_.type=col.type;
            }
        }
        is_desc_ = is_desc;
        tuple_num = 0;
        tuple_len_=prev_->tupleLen();
        used_tuple.clear();
        sort_mac=std::make_unique<sort_machine>(is_desc_,"sim_sort",prev_->tupleLen(),col_.offset,col_.type,col_.len,10000000);
    }

    void beginTuple() override { 
        prev_->beginTuple();
        while(!prev_->is_end()){
            auto rec=prev_->Next();
            sort_mac->write_rec(rec->data);
            prev_->nextTuple();
        }
        sort_mac->sort_it();
        std::ifstream outFile1("sim_sort.txt");
        readFile=std::move(outFile1);
        current_tuple=std::make_unique<RmRecord>(tuple_len_);
        readFile.read(current_tuple->data,tuple_len_);
        if(readFile.gcount()==0)isend=true;
    }

    void nextTuple() override {
        current_tuple=std::make_unique<RmRecord>(tuple_len_);
        readFile.read(current_tuple->data,tuple_len_);
        if(readFile.gcount()==0)isend=true;
    }

    std::unique_ptr<RmRecord> Next() override {
        
        
        return std::move(current_tuple);
    }

    Rid &rid() override { return _abstract_rid; }

    const std::vector<ColMeta> &cols() const override {
        return prev_->cols();
    };

    bool is_end()const override{return isend;}

    size_t tupleLen() const override{ return tuple_len_; };
};