/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */
#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class AggregationExecutor : public AbstractExecutor {
        public:
        std::unique_ptr<AbstractExecutor> prev_;
        std::vector<TabCol> group_cols_;
        std::vector<ColMeta> cols_;
        std::vector<ColMeta> having_cols_;
        std::vector<Condition> having_conds_;
        std::map<std::vector<char>,std::vector<RmRecord>> mp;
        std::map<std::vector<char>, std::vector<RmRecord>>::iterator point=mp.end();
        std::vector<std::pair<int,int>> sel_col_offset_and_size;
        std::vector<std::pair<int,int>> have_col_offset_and_size;
        bool special_judge_for_count_when_no_column_feats=true;
        int col_len=0;
        AggregationExecutor(std::unique_ptr<AbstractExecutor> prev,std::vector<TabCol> cols,std::vector<TabCol> group_cols, std::vector<Condition> having_conds) {
                prev_=std::move(prev);
                group_cols_=std::move(group_cols);
                having_conds_=std::move(having_conds);
                int offset_=0;
                for(auto col:cols){
                        ColMeta meta=ColMeta();
                        meta.aggr_type=col.aggr_type;
                        meta.name=col.col_name;
                        meta.new_name=col.new_col_name;
                        meta.offset=offset_;
                        if(meta.aggr_type==ast::AggrType::COUNT){
                                meta.len=4;
                                offset_+=4;
                                col_len+=4;
                                meta.type=ColType::TYPE_INT;
                        }
                        else{
                                special_judge_for_count_when_no_column_feats=0;
                                int col_len_=0;
                                ColType col_type;
                                for(auto prev_col:prev_->cols()){
                                        if(prev_col.tab_name==col.tab_name&&prev_col.name==col.col_name){
                                                col_len_=prev_col.len;
                                                col_type=prev_col.type;
                                                break;
                                        }
                                }
                                assert(col_len_>0);
                                meta.len=col_len_;
                                offset_+=col_len_;
                                col_len+=col_len_;
                                meta.type=col_type;
                        }
                        meta.tab_name=col.tab_name;
                        
                        
                        cols_.push_back(meta);
                }

                for(auto cond:having_conds_){
                        auto col=cond.lhs_col;
                        ColMeta meta=ColMeta();
                        meta.aggr_type=col.aggr_type;
                        meta.name=col.col_name;
                        if(meta.name=="*"){
                                meta.name=cols_[0].name;
                                meta.tab_name=cols_[0].tab_name;
                        }
                        else{
                                meta.tab_name=col.tab_name;;
                        }
                        if(col.aggr_type==ast::AggrType::COUNT){
                                meta.len=4;
                                meta.type=ColType::TYPE_INT;
                        }
                        else{
                                int col_len_=0;
                                ColType col_type;
                                for(auto prev_col:prev_->cols()){
                                        if(prev_col.tab_name==col.tab_name&&prev_col.name==col.col_name){
                                                col_len_=prev_col.len;
                                                col_type=prev_col.type;
                                                break;
                                        }
                                }
                                meta.len=col_len_;
                                meta.type=col_type;
                        }
                        having_cols_.push_back(meta);
                }
        }

        void beginTuple() override {//lots to do here
                std::vector<std::pair<int,int>> group_col_offset_and_size;

                prev_->beginTuple();
                if(prev_->is_end()){
                        return;
                }
                special_judge_for_count_when_no_column_feats=0;
                for(auto col:prev_->cols()){
                        for(auto group_col:group_cols_){
                                if(col.name==group_col.col_name&&col.tab_name==group_col.tab_name){
                                        group_col_offset_and_size.emplace_back(col.offset,col.len);
                                }
                        }
                }
                for(auto sel_col:cols_){
                        for(auto col:prev_->cols()){
                                if(col.name==sel_col.name&&col.tab_name==sel_col.tab_name){
                                        sel_col_offset_and_size.emplace_back(col.offset,col.len);
                                }
                                if(sel_col.aggr_type==ast::AggrType::NO_AGGR&&(group_cols_.size()>1
                                ||group_cols_[0].col_name!=sel_col.name||group_cols_[0].tab_name!=sel_col.tab_name)){
                                        throw IllegalGroupByError();
                                }
                        }
                }
                for(auto have_col:having_cols_){
                        for(auto col:prev_->cols()){
                                if(col.name==have_col.name&&col.tab_name==have_col.tab_name){
                                        have_col_offset_and_size.emplace_back(col.offset,col.len);
                                }
                                if(have_col.aggr_type==ast::AggrType::NO_AGGR&&(group_cols_.size()>1
                                ||group_cols_[0].col_name!=have_col.name||group_cols_[0].tab_name!=have_col.tab_name)){
                                        throw IllegalGroupByError();
                                }
                        }
                }
                while(!prev_->is_end()){
                        auto res=prev_->Next();
                        std::vector<char> group_value;
                        for(const auto&pr:group_col_offset_and_size){
                                for(int i=0;i<pr.second;i++){
                                        group_value.push_back(*(res->data+pr.first+i));
                                }
                        }
                        //if(mp.size()<1000000)
                        mp[group_value].push_back(*res.get());
                        prev_->nextTuple();
                }
                point=mp.begin();
                while(point!=mp.end()&&!check())point++;
        }

        void nextTuple() override {
                if(point==mp.end())return;
                point++;
                while(point!=mp.end()&&!check())point++;
        }

        std::unique_ptr<RmRecord> Next() override {
                
                if(point==mp.end()&&special_judge_for_count_when_no_column_feats){
                        special_judge_for_count_when_no_column_feats=false;
                        auto res=std::make_unique<RmRecord>(col_len);
                        for(int i=0;i<col_len/4;i++){
                                *(int*)(res->data+i*4)=0;
                        }
                        return res;
                }
                if(is_end())return nullptr;
                special_judge_for_count_when_no_column_feats=false;

                auto target_rec=std::make_unique<RmRecord>(col_len);
                std::vector<Value> v(sel_col_offset_and_size.size());

                auto first_res=*(*point).second.begin();
                for(int i=0;i<sel_col_offset_and_size.size();i++){
                        if(cols_[i].aggr_type==ast::AggrType::COUNT){
                                v[i].int_val=1;
                                continue;
                        }
                        if(cols_[i].type==ColType::TYPE_INT){
                                v[i].set_int(*(int*)(first_res.data+sel_col_offset_and_size[i].first));
                        }
                        else if(cols_[i].type==ColType::TYPE_FLOAT){
                                v[i].set_float(*(float*)(first_res.data+sel_col_offset_and_size[i].first));
                        }
                        else if(cols_[i].type==ColType::TYPE_STRING){
                                std::string s=std::string((char *)first_res.data+sel_col_offset_and_size[i].first,sel_col_offset_and_size[i].second);
                                v[i].set_str(s);
                        }
                }
                bool first_one=true;
                for(const auto& record:(*point).second){
                        if(first_one){
                                first_one=false;
                                continue;
                        }
                        for(int i=0;i<cols_.size();i++){
                                if(cols_[i].aggr_type==ast::AggrType::COUNT){
                                        v[i].int_val++;
                                        continue;
                                }
                                if(cols_[i].type==ColType::TYPE_INT){
                                        int target_int=*(int*)(record.data+sel_col_offset_and_size[i].first);
                                        if(cols_[i].aggr_type==ast::AggrType::MAX){
                                                v[i].int_val=std::max(v[i].int_val,target_int);
                                        }
                                        else if(cols_[i].aggr_type==ast::AggrType::MIN){
                                                v[i].int_val=std::min(v[i].int_val,target_int);
                                        }
                                        else if(cols_[i].aggr_type==ast::AggrType::SUM){
                                                v[i].int_val+=target_int;
                                        }
                                        else{
                                                v[i].int_val=target_int;
                                        }

                                }
                                else if(cols_[i].type==ColType::TYPE_FLOAT){
                                        float target_float=*(float*)(record.data+sel_col_offset_and_size[i].first);
                                        if(cols_[i].aggr_type==ast::AggrType::MAX){
                                                v[i].float_val=std::max(v[i].float_val,target_float);
                                        }
                                        else if(cols_[i].aggr_type==ast::AggrType::MIN){
                                                v[i].float_val=std::min(v[i].float_val,target_float);
                                        }
                                        else if(cols_[i].aggr_type==ast::AggrType::SUM){
                                                v[i].float_val+=target_float;
                                        }
                                        else{
                                                v[i].float_val=target_float;
                                        }
                                }
                        }
                }

                for(int i=0;i<cols_.size();i++){
                        v[i].init_raw(cols_[i].len);
                        memcpy(target_rec->data+cols_[i].offset,v[i].raw->data,v[i].raw->size);
                }
                return std::move(target_rec);
        }

        Rid &rid() override { return _abstract_rid; }

        bool is_end()const override{return special_judge_for_count_when_no_column_feats==0&&point==mp.end();}

        const std::vector<ColMeta> &cols() const override{
                return cols_;
        };

        bool check(){
                auto first_res=*(*point).second.begin();

                std::vector<Value> v(have_col_offset_and_size.size());

                for(int i=0;i<have_col_offset_and_size.size();i++){
                        if(having_cols_[i].aggr_type==ast::AggrType::COUNT){
                                v[i].int_val=1;
                                continue;
                        }
                        if(having_cols_[i].type==ColType::TYPE_INT){
                                v[i].set_int(*(int*)(first_res.data+have_col_offset_and_size[i].first));
                        }
                        else if(having_cols_[i].type==ColType::TYPE_FLOAT){
                                v[i].set_float(*(float*)(first_res.data+have_col_offset_and_size[i].first));
                        }
                        else if(having_cols_[i].type==ColType::TYPE_STRING){
                                std::string s=std::string((char *)first_res.data+have_col_offset_and_size[i].first,have_col_offset_and_size[i].second);
                                v[i].set_str(s);
                        }
                }
                bool first_one=true;
                for(const auto& record:(*point).second){
                        if(first_one){
                                first_one=false;
                                continue;
                        }
                        for(int i=0;i<having_cols_.size();i++){
                                if(having_cols_[i].aggr_type==ast::AggrType::COUNT){
                                        v[i].int_val++;
                                        continue;
                                }
                                if(having_cols_[i].type==ColType::TYPE_INT){
                                        int target_int=*(int*)(record.data+have_col_offset_and_size[i].first);
                                        if(having_cols_[i].aggr_type==ast::AggrType::MAX){
                                                v[i].int_val=std::max(v[i].int_val,target_int);
                                        }
                                        else if(having_cols_[i].aggr_type==ast::AggrType::MIN){
                                                v[i].int_val=std::min(v[i].int_val,target_int);
                                        }
                                        else if(having_cols_[i].aggr_type==ast::AggrType::SUM){
                                                v[i].int_val+=target_int;
                                        }
                                        else{
                                                v[i].int_val=target_int;
                                        }

                                }
                                else if(having_cols_[i].type==ColType::TYPE_FLOAT){
                                        float target_float=*(float*)(record.data+have_col_offset_and_size[i].first);
                                        if(having_cols_[i].aggr_type==ast::AggrType::MAX){
                                                v[i].float_val=std::max(v[i].float_val,target_float);
                                        }
                                        else if(having_cols_[i].aggr_type==ast::AggrType::MIN){
                                                v[i].float_val=std::min(v[i].float_val,target_float);
                                        }
                                        else if(having_cols_[i].aggr_type==ast::AggrType::SUM){
                                                v[i].float_val+=target_float;
                                        }
                                        else{
                                                v[i].float_val=target_float;
                                        }
                                }
                        }
                }
                for(int i=0;i<having_cols_.size();i++){
                        v[i].init_raw(having_cols_[i].len);
                        std::vector<ColType> tmpColV;
                        std::vector<int> tmpIntV;
                        tmpColV.push_back(v[i].type);
                        tmpIntV.push_back(v[i].raw->size);
                        auto cond=having_conds_[i];

                        if(cond.op==CompOp::OP_IN){
                                bool found_one=false;
                                for(auto &r_rec:*(cond.rhs_val.vec_)){
                                        if(ix_compare(v[i].raw->data,r_rec.raw->data,tmpColV,tmpIntV)==0){
                                                found_one=true;
                                        }
                                }
                                if(!found_one)return false;
                                continue;

                        }

                        auto res=ix_compare(v[i].raw->data,cond.rhs_val.raw->data,tmpColV,tmpIntV);
                        if (cond.op == CompOp::OP_EQ) {
                                if(res)return false;
                        } else if (cond.op == CompOp::OP_NE) {
                                if(res==0)return false;
                        } else if (cond.op == CompOp::OP_LT) {
                                if(res>=0)return false;
                        } else if (cond.op == CompOp::OP_GT) {
                                if(res<=0)return false;
                        } else if (cond.op == CompOp::OP_LE) {
                                if(res>0)return false;
                        } else if (cond.op == CompOp::OP_GE) {
                                if(res<0)return false;
                        }else {
                                throw InternalError("Invalid operation");
                                return false;
                        }
                }
                return true;
        }

};