/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class DeleteExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;                   // 表的元数据
    std::vector<Condition> conds_;  // delete的条件
    RmFileHandle *fh_;              // 表的数据文件句柄
    std::vector<Rid> rids_;         // 需要删除的记录的位置
    std::string tab_name_;          // 表名称
    SmManager *sm_manager_;

   public:
    DeleteExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<Condition> conds,
                   std::vector<Rid> rids, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        context_ = context;
    }

    std::unique_ptr<RmRecord> Next() override {
        //context_->lock_mgr_->lock_IX_on_table(context_->txn_,sm_manager_->fhs_[tab_name_].get()->GetFd());
        context_->lock_mgr_->lock_IX_on_table(context_->txn_,fh_->GetFd());
        for(const auto &rid:rids_){
            auto& indexes=tab_.indexes;
            const auto& rec=sm_manager_->fhs_[tab_name_]->get_record(rid,context_);
            sm_manager_->fhs_[tab_name_]->delete_record(rid,context_);
            WriteRecord* writerecord=new WriteRecord(WType::DELETE_TUPLE,tab_name_,rid,*rec.get());
            context_->txn_->append_write_record(writerecord);
            for(auto idx:indexes){
                const auto &idx_name=sm_manager_->get_ix_manager()->get_index_name(tab_name_,idx.cols);
                auto idx_entry=sm_manager_->ihs_[idx_name].get();
                char* key=new char[idx.col_tot_len];
                int cur_offset=0;
                for(auto col:idx.cols){
                    memcpy(key+cur_offset,rec->data+col.offset,col.len);
                    cur_offset+=col.len;
                }
                bool res=idx_entry->delete_entry(key,context_->txn_);    
                assert(res);
                delete[]key;
            }
            auto log_record=new DeleteLogRecord(context_->txn_->get_transaction_id(),*rec,rid,tab_name_);
            context_->log_mgr_->add_log_to_buffer(log_record);
        }
        return nullptr;
    }

    Rid &rid() override { return _abstract_rid; }
};