/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class ProjectionExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> prev_;        // 投影节点的儿子节点
    std::vector<ColMeta> cols_;                     // 需要投影的字段
    size_t len_;                                    // 字段总长度
    std::vector<size_t> sel_idxs_;                  

   public:
    ProjectionExecutor(std::unique_ptr<AbstractExecutor> prev, const std::vector<TabCol> &sel_cols) {
        prev_ = std::move(prev);
        size_t curr_offset = 0;
        auto prev_cols = prev_->cols();
        for (auto &sel_col : sel_cols) {
            std::vector<ColMeta>::iterator pos;
            for(pos=prev_cols.begin();pos!=prev_cols.end();pos++){
                if(pos->name==sel_col.col_name&&(pos->tab_name==sel_col.tab_name)&&pos->new_name==sel_col.new_col_name){
                    break;
                }
            }
            sel_idxs_.push_back(pos->offset);
            auto col = *pos;
            col.offset = curr_offset;
            curr_offset += col.len;
            cols_.push_back(col);
        }
        len_ = curr_offset;
    }

    void beginTuple() override {prev_->beginTuple();}

    void nextTuple() override {prev_->nextTuple();}

    std::unique_ptr<RmRecord> Next() override {
        auto res=prev_->Next();
        auto new_rec=std::make_unique<RmRecord>(len_);
        if(res==nullptr)return nullptr;
        int cnt=0;
        for(const auto&col:cols_){
            memcpy(new_rec->data+col.offset,res->data+sel_idxs_[cnt++],col.len);
        }

        return new_rec;
    }

    Rid &rid() override { return _abstract_rid; }

    bool is_end()const override{return prev_->is_end();}

    const std::vector<ColMeta> &cols() const override{
        return cols_;
    };
};