/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class UpdateExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;
    std::vector<Condition> conds_;
    RmFileHandle *fh_;
    std::vector<Rid> rids_;
    std::string tab_name_;
    std::vector<SetClause> set_clauses_;
    SmManager *sm_manager_;

   public:
    UpdateExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<SetClause> set_clauses,
                   std::vector<Condition> conds, std::vector<Rid> rids, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        set_clauses_ = set_clauses;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        context_ = context;
    }
    std::unique_ptr<RmRecord> Next() override {
        //context_->lock_mgr_->lock_IX_on_table(context_->txn_,sm_manager_->fhs_[tab_name_].get()->GetFd());
        context_->lock_mgr_->lock_IX_on_table(context_->txn_,fh_->GetFd());
        std::vector<std::vector<char*>> potential_keys(tab_.indexes.size());

        std::vector<std::vector<char*>> old_keys(tab_.indexes.size());

        for(const auto&rid:rids_){
            auto& indexes=tab_.indexes;
            auto new_rec=fh_->get_record(rid,context_);
            char* old_rec=new char[new_rec->size];
            memcpy(old_rec,new_rec->data,new_rec->size);
            int idx_idx=0;
            for(auto idx:indexes){
                const auto &idx_name=sm_manager_->get_ix_manager()->get_index_name(tab_name_,idx.cols);
                auto idx_entry=sm_manager_->ihs_[idx_name].get();
                char* key=new char[idx.col_tot_len];
                int cur_offset=0;
                for(auto col:idx.cols){
                    memcpy(key+cur_offset,old_rec+col.offset,col.len);
                    cur_offset+=col.len;
                }
                old_keys[idx_idx].push_back(key);
                idx_idx++;
            }
            delete[]old_rec;
        }

        for(const auto&rid:rids_){
            auto& indexes=tab_.indexes;
            auto new_rec=fh_->get_record(rid,context_);
            char* test_rec=new char[new_rec->size];
            memcpy(test_rec,new_rec->data,new_rec->size);
            for(auto pr:set_clauses_){
                auto col=tab_.cols.begin();
                for(;col!=tab_.cols.end();col++){
                    if(col->name==pr.lhs.col_name)break;
                }
                pr.rhs.raw=nullptr;
                pr.rhs.init_raw(col->len);
                if(!pr.pre_add)memcpy(test_rec+col->offset,pr.rhs.raw->data,col->len);
                else{
                    if(col->type==ColType::TYPE_INT){
                        *(test_rec+col->offset)=*(int*)(test_rec+col->offset)+*(int*)(pr.rhs.raw->data);
                    }
                    else if(col->type==ColType::TYPE_FLOAT){
                        *(test_rec+col->offset)=*(float*)(test_rec+col->offset)+*(float*)(pr.rhs.raw->data);
                    }
                    else throw InternalError("pre add update can't be implied on string");
                }
            }

            
            int idx_idx=0;
            for(auto idx:indexes){
                const auto &idx_name=sm_manager_->get_ix_manager()->get_index_name(tab_name_,idx.cols);
                auto idx_entry=sm_manager_->ihs_[idx_name].get();
                char* key=new char[idx.col_tot_len];
                int cur_offset=0;
                for(auto col:idx.cols){
                    memcpy(key+cur_offset,test_rec+col.offset,col.len);
                    cur_offset+=col.len;
                }
                std::vector<ColType>col_types;
                std::vector<int>col_lens;
                for(auto col:idx.cols){
                    col_types.push_back(col.type);
                    col_lens.push_back(col.len);
                }
                
                // auto res=idx_entry->lower_bound(key);
                // Rid res_rid;
                // if(res==idx_entry->leaf_end()){
                //     res_rid.page_no=-1;//means the end of the b_plus_tree
                //     res_rid.slot_no=-1;
                // }
                // else{
                //     res_rid=idx_entry->get_rid(res);
                // }
                // context_->lock_mgr_->lock_exclusive_gap(context_->txn_,res_rid,idx_name);
                
                

                bool exsits_in_old_key=false;
                for(auto old_key:old_keys[idx_idx]){
                    if(ix_compare(key,old_key,col_types,col_lens)==0){
                        exsits_in_old_key=true;
                        break;
                    }
                }
                if(!exsits_in_old_key){
                    if(idx_entry->get_value(key,context_->txn_)){
                        std::string tmp="failure\n";
                        write_output(tmp,context_);
                        return nullptr;
                    }
                }
                for(auto po_key:potential_keys[idx_idx]){
                    if(ix_compare(po_key,key,col_types,col_lens)==0){
                        std::string tmp="failure\n";
                        write_output(tmp,context_);
                        return nullptr;
                    }
                }
                potential_keys[idx_idx].push_back(key);
                idx_idx++;
            }
            delete[]test_rec;
        }

        for(auto &tmp:old_keys){
            for(auto &chr:tmp){
                delete[]chr;
            }
        }
        for(auto &tmp:potential_keys){
            for(auto &chr:tmp){
                delete[]chr;
            }
        }


        for(const auto&rid:rids_){
            auto& indexes=tab_.indexes;
            auto new_rec=fh_->get_record(rid,context_);
            RmRecord old_rec=*new_rec;
            char* old_rec_value=new char[new_rec->size];
            memcpy(old_rec_value,new_rec->data,new_rec->size);
            WriteRecord* writerecord=new WriteRecord(WType::UPDATE_TUPLE,tab_name_,rid,*new_rec.get());
            for(auto &pr:set_clauses_){
                auto col=tab_.cols.begin();
                for(;col!=tab_.cols.end();col++){
                    if(col->name==pr.lhs.col_name)break;
                }
                pr.rhs.raw=nullptr;
                pr.rhs.init_raw(col->len);
                if(!pr.pre_add)memcpy(new_rec->data+col->offset,pr.rhs.raw->data,col->len);
                else{
                    if(col->type==ColType::TYPE_INT){
                        *(new_rec->data+col->offset)=*(int*)(new_rec->data+col->offset)+*(int*)(pr.rhs.raw->data);
                    }
                    else if(col->type==ColType::TYPE_FLOAT){
                        *(new_rec->data+col->offset)=*(float*)(new_rec->data+col->offset)+*(float*)(pr.rhs.raw->data);
                    }
                    else throw InternalError("pre add update can't be implied on string");
                }
            }
            fh_->update_record(rid,new_rec->data,context_);

            auto log_record=new UpdateLogRecord(context_->txn_->get_transaction_id(),old_rec,*new_rec,rid,tab_name_);
            context_->log_mgr_->add_log_to_buffer(log_record);

            context_->txn_->append_write_record(writerecord);
            for(auto idx:indexes){
                const auto &idx_name=sm_manager_->get_ix_manager()->get_index_name(tab_name_,idx.cols);
                auto idx_entry=sm_manager_->ihs_[idx_name].get();
                char* old_key=new char[idx.col_tot_len];
                int cur_offset=0;
                for(auto col:idx.cols){
                    memcpy(old_key+cur_offset,old_rec_value+col.offset,col.len);
                    cur_offset+=col.len;
                }
                //context_->lock_mgr_->lock_exclusive_gap(context_->txn_,rid,idx_name);
                

                char* key=new char[idx.col_tot_len];
                cur_offset=0;
                for(auto col:idx.cols){
                    memcpy(key+cur_offset,new_rec->data+col.offset,col.len);
                    cur_offset+=col.len;
                }
                std::vector<ColType>col_types;
                std::vector<int>col_lens;
                for(auto col:idx.cols){
                    col_types.push_back(col.type);
                    col_lens.push_back(col.len);
                }

                if(ix_compare(old_key,key,col_types,col_lens)==0){
                    delete[]old_key;
                    delete[]key;
                    continue;
                }

                bool res=idx_entry->delete_entry(old_key,context_->txn_);
                assert(res);
                delete[]old_key;
                res=idx_entry->insert_entry(key,rid,context_->txn_);
                assert(res);
                delete[]key;
            }
            delete[]old_rec_value;
        }
        return nullptr;
    }

    Rid &rid() override { return _abstract_rid; }
};