#pragma once
#include<fstream>
#include<vector>
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include <iostream>
#include <sstream>
#include <ctime>
#include <cstdlib>

struct sort_node{
    char* val;
    int offset=0;
    ColType type;
    int len;
    int idx;
    bool is_desc;
    sort_node(char* val_,int offset_,ColType type_,int len_,bool is_desc_,int idx_=0){
        val=val_;
        offset=offset_;
        type=type_;
        len=len_;
        is_desc=is_desc_;
        idx=idx_;
    }
    bool operator<(const sort_node& b)const   //重载<运算符
	{
		if(!is_desc) return ix_compare(val+offset,b.val+b.offset,type,len)<0;
        else return ix_compare(val+offset,b.val+b.offset,type,len)>0;
	}
};


class sort_machine{
    private:
        std::ofstream ofs;
        std::vector<std::string> file_names;
        std::vector<sort_node> cur_recs;
        int file_num=0;
        int file_max_len=4000;
        int cur_len=0;
        bool desc=false;
        int target_offset=0;
        ColType target_type;
        int target_len;
        std::string main_file_name;
        int record_len;



        // const static inline int sort_node_ix_compare(const char *a, const char *b, ColType type, int col_len) {
        //     switch (type) {
        //         case TYPE_INT: {
        //             int ia = *(int *)a;
        //             int ib = *(int *)b;
        //             return (ia < ib) ? -1 : ((ia > ib) ? 1 : 0);
        //         }
        //         case TYPE_FLOAT: {
        //             float fa = *(float *)a;
        //             float fb = *(float *)b;
        //             return (fa < fb) ? -1 : ((fa > fb) ? 1 : 0);
        //         }
        //         case TYPE_STRING:
        //             return memcmp(a, b, col_len);
        //         default:
        //             throw InternalError("Unexpected data type");
        //     }
        // }


        // const static bool cmp_sort_node(const sort_node & a,const sort_node & b){
        //     if(!a.is_desc) return sort_node_ix_compare(a.val+a.offset,b.val+b.offset,a.type,a.len)<0;
        //     else return sort_node_ix_compare(a.val+a.offset,b.val+b.offset,a.type,a.len)>0;
        // }

    public:
    sort_machine(bool desc_,std::string main_file_name_,int record_len_,int target_offset_,ColType target_type_,int target_len_,int k){
        file_num=file_names.size();
        desc=desc_;
        main_file_name=main_file_name_;
        record_len=record_len_;
        file_max_len=std::max(k,record_len)/record_len;
        target_offset=target_offset_;
        target_type=target_type_;
        target_len=target_len_;
    }
    void write_rec(char* rec){
        char* rec1=(char*)malloc(record_len);
        memcpy(rec1,rec,record_len);
        if(cur_len>=file_max_len){
            std::sort(cur_recs.begin(),cur_recs.end());
            auto new_file_name=main_file_name+std::to_string(file_num)+".txt";
            std::ofstream outFile(new_file_name);
            for(const auto&node:cur_recs){
                outFile.write(node.val,record_len);
                free(node.val);
            }
            outFile.close();
            file_names.push_back(new_file_name);
            cur_recs.clear();
            cur_len=0;
            file_num++;
        }
        cur_len++;
        auto node=new sort_node(rec1,target_offset,target_type,target_len,desc);
        cur_recs.push_back(*node);
    }
    void sort_it(){
        if(cur_len){
            std::sort(cur_recs.begin(),cur_recs.end());
            auto new_file_name=main_file_name+std::to_string(file_num)+".txt";
            std::ofstream outFile(new_file_name);
            for(const auto&node:cur_recs){
                outFile.write(node.val,record_len);
                free(node.val);
            }
            outFile.close();
            file_names.push_back(new_file_name);
            cur_recs.clear();
            cur_len=0;
            file_num++;
        }
        std::vector<std::ifstream> opening_files(file_num);
        std::set<sort_node> val_and_filenum_st;
        std::vector<char*>files_cur_recs(file_num);
        //td::vector<>
        for(int i=0;i<file_num;i++){
            std::ifstream inFile(file_names[i]);
            char* res=(char*)malloc(record_len);
            
            inFile.read(res,record_len);
            if(inFile.gcount()==0){
                opening_files[i]=std::move(inFile);
                continue;
            }
            auto node=new sort_node(res,target_offset,target_type,target_len,desc,i);
            val_and_filenum_st.insert(*node);
            opening_files[i]=std::move(inFile);
        }
        std::ofstream outFile(main_file_name+".txt");
        while(!val_and_filenum_st.empty()){
            char* now=val_and_filenum_st.begin()->val;
            int idx=val_and_filenum_st.begin()->idx;
            val_and_filenum_st.erase(val_and_filenum_st.begin());
            outFile.write(now,record_len);
            opening_files[idx].read(now,record_len);
            if(opening_files[idx].gcount()==0){
                free(now);
                continue;
            }
            auto node=new sort_node(now,target_offset,target_type,target_len,desc,idx);
            val_and_filenum_st.emplace(*node);
        }       
        for(int i=0;i<file_num;i++){
            opening_files[i].close();
        }
        outFile.close();
    }
    
};