/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "log_recovery.h"

/**
 * @description: analyze阶段，需要获得脏页表（DPT）和未完成的事务列表（ATT）
 */
void RecoveryManager::analyze() {
        int offset=0;
        char* log_header=(char*)malloc(LOG_HEADER_SIZE);
        while(disk_manager_->read_log(log_header, LOG_HEADER_SIZE, offset)){
                auto log_type = *(LogType*)(log_header);
                auto lsn=*(lsn_t*)(log_header+OFFSET_LSN);
                auto log_tot_len=*(uint32_t*)(log_header+OFFSET_LOG_TOT_LEN);
                auto log_tid=*(txn_id_t*)(log_header+OFFSET_LOG_TID);
                max_txn_id=std::max(log_tid,max_txn_id);
                auto prev_lsn=*(lsn_t*)(log_header+OFFSET_PREV_LSN);
                
                if(log_type!=LogType::END&&!xact_table.count(log_tid)){
                        xact_table[log_tid]={TranscationRecoveStatus::RUNING,lsn};
                }
                if(log_type==LogType::ABORT){
                        xact_table[log_tid].first=TranscationRecoveStatus::ABORTING;
                        auto log_record=new AbortLogRecord();
                        log_record->deserialize(log_header);
                        log_records.push_back(log_record);
                }
                else if(log_type==LogType::commit){
                        xact_table[log_tid].first=TranscationRecoveStatus::COMMITTING;
                        auto log_record=new CommitLogRecord();
                        log_record->deserialize(log_header);
                        log_records.push_back(log_record);
                }
                else if(log_type==LogType::END){
                        xact_table.erase(log_tid);
                        auto log_record=new EndLogRecord();
                        log_record->deserialize(log_header);
                        log_records.push_back(log_record);
                        
                }
                else if(log_type==LogType::begin){
                        
                }
                else{
                        char* log_buffer=(char*)malloc(log_tot_len);
                        disk_manager_->read_log(log_buffer,log_tot_len,offset);
                        if(log_type==LogType::INSERT){
                                auto log_record=new InsertLogRecord();
                                log_record->deserialize(log_buffer);
                                std::string table_name(log_record->table_name_,log_record->table_name_size_);
                                auto fh_=sm_manager_->fhs_.at(table_name).get();
                                PageId page_id;
                                page_id.fd=fh_->GetFd();
                                page_id.page_no=log_record->rid_.page_no; 
                                if(!DPT.count(page_id))DPT[page_id]=lsn;
                                log_record->prev_lsn_=xact_table[log_record->log_tid_].second;
                                xact_table[log_record->log_tid_].second=log_record->lsn_;
                                log_records.push_back(log_record);
                        }
                        else if(log_type==LogType::DELETE){
                                auto log_record=new DeleteLogRecord();
                                log_record->deserialize(log_buffer);
                                std::string table_name(log_record->table_name_,log_record->table_name_size_);
                                auto fh_=sm_manager_->fhs_.at(table_name).get();
                                PageId page_id;
                                page_id.fd=fh_->GetFd();
                                page_id.page_no=log_record->rid_.page_no;    
                                if(!DPT.count(page_id))DPT[page_id]=lsn;
                                log_record->prev_lsn_=xact_table[log_record->log_tid_].second;
                                xact_table[log_record->log_tid_].second=log_record->lsn_;

                                log_records.push_back(log_record);
                        }
                        else if(log_type==LogType::UPDATE){
                                auto log_record=new UpdateLogRecord();
                                log_record->deserialize(log_buffer);
                                std::string table_name(log_record->table_name_,log_record->table_name_size_);
                                auto fh_=sm_manager_->fhs_.at(table_name).get();
                                PageId page_id;
                                page_id.fd=fh_->GetFd();
                                page_id.page_no=log_record->rid_.page_no;    
                                if(!DPT.count(page_id))DPT[page_id]=lsn;
                                log_record->prev_lsn_=xact_table[log_record->log_tid_].second;
                                xact_table[log_record->log_tid_].second=log_record->lsn_;
                                log_records.push_back(log_record);
                        }
                        else assert(0);
                        free(log_buffer);

                }
                offset+=log_tot_len;
        }
        std::vector<txn_id_t> delete_lst;
        for(auto& txn_pr:xact_table){
                if(txn_pr.second.first==TranscationRecoveStatus::COMMITTING){
                        delete_lst.push_back(txn_pr.first);
                        auto rec=new EndLogRecord(txn_pr.first);
                        log_records.push_back(rec);
                }
        }
        for(auto &txn:delete_lst){
                xact_table.erase(txn);
        }
        free(log_header);
}

/**
 * @description: 重做所有未落盘的操作
 */
void RecoveryManager::redo() {
        for(auto log_record:log_records){
                if(log_record->log_type_==LogType::INSERT){
                        auto insert_log_record=(InsertLogRecord*)log_record;
                        std::string table_name(insert_log_record->table_name_,insert_log_record->table_name_size_);
                        auto fh_ = sm_manager_->fhs_.at(table_name).get();
                        PageId page_id;
                        page_id.fd=fh_->GetFd();
                        page_id.page_no=insert_log_record->rid_.page_no;    
                        // if(DPT.count(page_id)==0){
                        //         continue;//case 1
                        // }
                        // if(DPT[page_id]>log_record->lsn_)continue;//case 2
                        //auto page=buffer_pool_manager_->fetch_page(page_id);
                        // if (page->get_page_lsn()>log_record->lsn_) {//case 3
                        //         buffer_pool_manager_->unpin_page(page_id,false); 
                        //         continue;
                        // }
                        
                        sm_manager_->insert_for_recovery(table_name,(insert_log_record)->rid_,(insert_log_record)->insert_value_);
                        //buffer_pool_manager_->unpin_page(page_id,true);
                }
                else if(log_record->log_type_==LogType::DELETE){
                        auto delete_log_record=(DeleteLogRecord*)log_record;
                        std::string table_name(delete_log_record->table_name_,delete_log_record->table_name_size_);
                        auto fh_ = sm_manager_->fhs_.at(table_name).get();
                        PageId page_id;
                        page_id.fd=fh_->GetFd();
                        page_id.page_no=delete_log_record->rid_.page_no;    
                        // if(DPT.count(page_id)==0)continue;//case 1
                        // if(DPT[page_id]>log_record->lsn_)continue;//case 2
                        // auto page=buffer_pool_manager_->fetch_page(page_id);
                        // if (page->get_page_lsn()>log_record->lsn_) {//case 3
                        //         buffer_pool_manager_->unpin_page(page_id,false); 
                        //         continue;
                        // }
                        sm_manager_->delete_for_recovery(table_name,(delete_log_record)->rid_,(delete_log_record)->delete_value_);
                        //buffer_pool_manager_->unpin_page(page_id,true);
                }
                else if(log_record->log_type_==LogType::UPDATE){
                        auto update_log_record=(UpdateLogRecord*)log_record;
                        std::string table_name(update_log_record->table_name_,update_log_record->table_name_size_);
                        auto fh_ = sm_manager_->fhs_.at(table_name).get();
                        PageId page_id;
                        page_id.fd=fh_->GetFd();
                        page_id.page_no=update_log_record->rid_.page_no;    
                        // if(DPT.count(page_id)==0)continue;//case 1
                        // if(DPT[page_id]>log_record->lsn_)continue;//case 2
                        // auto page=buffer_pool_manager_->fetch_page(page_id);
                        // if (page->get_page_lsn()>log_record->lsn_) {//case 3
                        //         buffer_pool_manager_->unpin_page(page_id,false); 
                        //         continue;
                        // }
                        sm_manager_->update_for_recovery(table_name,(update_log_record)->rid_,(update_log_record)->old_value_,(update_log_record)->new_value_);
                        //buffer_pool_manager_->unpin_page(page_id,true);
                }
        }
}

/**
 * @description: 回滚未完成的事务
 */
void RecoveryManager::undo() {
        int len=log_records.size();
        for(int i=len-1;i>=0;i--){
                auto log_record=log_records[i];
                if(xact_table.count(log_record->log_tid_)==0)continue;
                if(log_record->log_type_==LogType::INSERT){
                        auto insert_log_record=(InsertLogRecord*)log_record;
                        std::string table_name(insert_log_record->table_name_,insert_log_record->table_name_size_);
                        sm_manager_->rollback_insert(table_name,(insert_log_record)->rid_,nullptr);
                }
                else if(log_record->log_type_==LogType::DELETE){
                        auto delete_log_record=(DeleteLogRecord*)log_record;
                        std::string table_name(delete_log_record->table_name_,delete_log_record->table_name_size_);
                        sm_manager_->rollback_delete(table_name,(delete_log_record)->rid_,(delete_log_record)->delete_value_,nullptr);
                }
                else if(log_record->log_type_==LogType::UPDATE){
                        auto update_log_record=(UpdateLogRecord*)log_record;
                        std::string table_name(update_log_record->table_name_,update_log_record->table_name_size_);
                        sm_manager_->rollback_update(table_name,(update_log_record)->rid_,(update_log_record)->old_value_,nullptr);
                }
        }
        //sm_manager_->flush_backup_idx();
}